#Importing Python Libraries
import sys
import numpy as np
import docplex.mp.model as cpx
import time
import math

class MWU:
    def __init__(self,v,e,k,file,eps):
        start=time.time()
        iota = 0.5
        gamma = eps
        eta = np.log(e)/gamma
        fgraph = open(file,"r")
        graph = fgraph.read().split("\n")[:-1]

        originalsources=[]
        #Segregating the respective values
        self.sourcesink = []
        self.demands = []
        self.weights = []
        self.edges=[]
        self.capacities={}
        self.v=v
        self.e=e
        self.k=k
        
        
        
        for c in range(len(graph)):
            temp = graph[c].split(",")
            if c<self.k:
                originalsources.append(int(temp[0]))
                self.sourcesink.append((c+self.v,int(temp[1])))
                self.demands.append(float(temp[2]))
                self.weights.append(float(temp[3]))
                self.edges.append((c+self.v,int(temp[0])))
                self.capacities[(c+self.v,int(temp[0]))]=float(temp[2])
                
            else:
                self.edges.append((int(temp[0]),int(temp[1])))
                self.edges.append((int(temp[1]),int(temp[0])))
                self.capacities[(int(temp[0]),int(temp[1]))]=float(temp[2])
                self.capacities[(int(temp[1]),int(temp[0]))]=float(temp[2])
        
        self.v+=k
        self.f_i_e_vars={}
        for e in self.edges:
            for i in range(self.k):
                self.f_i_e_vars[i,e]=0.0
        
        
        l_function={}
        flows={}
        for e in self.edges:
            l_function[e]=1.0
            flows[e]=0.0
            
        
        feasible=[True for i in range(k)]
        
        def printflows():
            for e in self.edges:
                print(str(e)+':'+str(flows[e]))
                
        iteration = 0
        istarprime=0
        istar=0
        rhoprime=[]
        for x in range(k-1):
            rhoprime.append(0)
        rhoprime=np.array(rhoprime)
        
        calculatemincostflows=True
        
        
        #Create MIP model
        opt_model = cpx.Model(name="MIP Model")
        
        # setting the constant edge capacities for the linear program
        l_e={}
        c_e={}
        for e in self.edges:
            l_e[e] = l_function[e]
            c_e[e] = self.capacities[e] 
        b_j = {i:0.0 for i in range(v)}
        
        # setting the variable flow values for the linear program
        f_e_vars = {e:opt_model.continuous_var(lb=0, name="f_{0}".format(e)) for e in self.edges}
        
        
        
        # adding the constraints for the linear program
        capacityConstraint = {e :
            opt_model.add_constraint(
                    ct=f_e_vars[e] <= c_e[e],
                    ctname="capacityConstraint_{0}".format(e))
                for e in self.edges}
        
        nonnegativeConstraint= {e :
        opt_model.add_constraint(
        ct=f_e_vars[e] >= 0 ,
        ctname="nonnegativeConstraint_{0}".format(e))
            for e in self.edges}
        
        for x in range(self.v):
            b_j[x]=0.0
        
        flowConservationConstraint = {j :
        opt_model.add_constraint(ct=opt_model.sum(f_e_vars[(j,a)] 
            for a in range(self.v) if (j,a) in self.edges) - opt_model.sum(f_e_vars[(a,j)]
            for a in range(self.v) if (a,j) in self.edges ) ==b_j[j],
            ctname="flowConservationConstraint_{0}".format(j))for j in range(self.v)}
            
        loopnumber = int((1-gamma)*10000)
        loopnumber = 10000
        #loopnumber = 8000
        #while True:
        for g in range(loopnumber):
            #if iteration % 10 == 0:
            print('it: '+str(iteration))       
            if calculatemincostflows:
                #print('flag')
                mincostflows=[]
                rho=[]
                for i in range(self.k):
                    #print(i)
                    if feasible[i]:
                        for e in self.edges:
                            l_e[e] = l_function[e]
                        #Remove previous source sink demands
                        opt_model.remove_constraint("flowConservationConstraint_"+str(self.sourcesink[i][0]))
                        opt_model.remove_constraint("flowConservationConstraint_"+str(self.sourcesink[i][1]))
                        #Update Source Sink Demands
                        b_j[self.sourcesink[i][0]]=self.demands[i]
                        b_j[self.sourcesink[i][1]]=-1*self.demands[i]
                        #Update Model
                        flowConservationConstraint = {j :
                            opt_model.add_constraint(ct=opt_model.sum(f_e_vars[(j,a)] 
                            for a in range(self.v) if (j,a) in self.edges) - opt_model.sum(f_e_vars[(a,j)]
                            for a in range(self.v) if (a,j) in self.edges ) ==b_j[j],
                            ctname="flowConservationConstraint_{0}".format(j))for j in [self.sourcesink[i][0],self.sourcesink[i][1]]}
                        
                        objective = opt_model.sum(l_e[e]*f_e_vars[e] for e in self.edges)
                        opt_model.minimize(objective)
                        
                        
                        try:
                            opt_model.solve()

                        except:
                            feasible[i]=False
                            rho.append(9999999999)
                            continue
                        if not feasible[i]:
                            print('error')
                        mincostflow={}
                        for e in self.edges:
                            mincostflow[e]=f_e_vars[e].solution_value
                        mincostflows.append(mincostflow)

                        summation=0
                        for e in self.edges:
                            summation+=(l_function[e]*mincostflow[e])/self.weights[i]
                        rho.append(summation)
                        #Replace Source Sink Demands with Zero
                        opt_model.remove_constraint("flowConservationConstraint_"+str(self.sourcesink[i][0]))
                        opt_model.remove_constraint("flowConservationConstraint_"+str(self.sourcesink[i][1]))
                        b_j[self.sourcesink[i][0]]=0.0
                        b_j[self.sourcesink[i][1]]=0.0
                        flowConservationConstraint = {j :
                            opt_model.add_constraint(ct=opt_model.sum(f_e_vars[(j,a)] 
                            for a in range(self.v) if (j,a) in self.edges) - opt_model.sum(f_e_vars[(a,j)]
                            for a in range(self.v) if (a,j) in self.edges ) ==b_j[j],
                            ctname="flowConservationConstraint_{0}".format(j))for j in [self.sourcesink[i][0],self.sourcesink[i][1]]}
                            
                    else:
                        rho.append(9999999999)
            rho = np.array(rho)
            istar=np.argmin(rho)
            y=0
            for x in range(self.k):
                if x != istar:
                    rhoprime[y]=rho[x]
                    y+=1
            istarprime=np.argmin(rhoprime)
            if np.amin(rho) == 9999999999:
                printflows()
                print('STOP2')
                end=time.time()
                print("Time taken = ",end-start," seconds")
                return
            delta=float('inf')
            for e in self.edges:
                #print('length '+str(len(mincostflows)))
                
                if mincostflows[istar][e]!=0:
                    temp = (gamma/eta)*(mincostflows[istar][e]/self.capacities[e])
                    if temp <= delta:
                        delta = temp
            tempflow = flows.copy()
            for e in self.edges:
                #print(mincostflows[istar][(a,b)])
                tempflow[e]=tempflow[e]+delta*mincostflows[istar][e]
                if tempflow[e]>self.capacities[e] or g ==loopnumber-1:

                    print('flow '+str(tempflow[e]))
                    print('edge '+str(e)+' '+str(self.capacities[e]))
                    print('STOP')
                    summation=0
                    for x in range(len(self.sourcesink)):
                        summation+=flows[(self.sourcesink[x][0],originalsources[x])]*(self.weights[x]/self.demands[x])
                    print('Max ANF Objective Value: '+str(summation))       
                    end=time.time()
                    print("Time taken = ",end-start," seconds")
                    return
                else:

                    l_function[e]=math.exp((eta*tempflow[e])/self.capacities[e])
            for e in self.edges:
                self.f_i_e_vars[istar,e]+=delta*mincostflows[istar][e]*(1/self.demands[i])
            flows=tempflow.copy()
            self.summation=0
            for e in self.edges:
                self.summation+=(l_function[e]*mincostflows[istar][e])/self.weights[istar]
            if self.summation <= rho[istar]*(1+iota):
                calculatemincostflows=False
            else:
                calculatemincostflows=True
            
            iteration+=1
            self.f_i_vars={}
            self.objective=0
            for x in range(len(self.sourcesink)):
                self.objective+=flows[(self.sourcesink[x][0],originalsources[x])]*(self.weights[x]/self.demands[x])
                self.f_i_vars[x]=flows[(self.sourcesink[x][0],originalsources[x])]*(1/self.demands[x])
            
        return None
        
    def getValues(self):
        f_i_e_vars_updated={}
        capacities_updated={}
        edges_updated=[]


                
        for e in self.edges:
            for i in range(self.k):
                if e[0]<(self.v-self.k) and e[1]<(self.v-self.k):
                    f_i_e_vars_updated[i,e]=self.f_i_e_vars[i,e]
                
       
 
        for e in self.edges:
            if e[0]<(self.v-self.k) and e[1]<(self.v-self.k):
                capacities_updated[e]=self.capacities[e]
                edges_updated.append(e)
        for e in edges_updated:
            flow=0
            for i in range(self.k):
                flow+=f_i_e_vars_updated[i,e]
                

        for e in edges_updated:
            f_e=0.0
            for i in range(self.k):
                f_e+=f_i_e_vars_updated[i,e]*(self.demands[i])
            if f_e >capacities_updated[e]:
                print('ANF CONSTRAINT VIOLATION')
                print(f_e)
                print(capacities_updated[e])
               
            
        return (self.objective,self.f_i_vars,f_i_e_vars_updated,edges_updated,capacities_updated,self.weights,self.demands,self.sourcesink)
                
        

            
                
        
            
            

  
    
    
